/*
 * $Id: buffer.c,v 1.12 2003/03/12 09:48:55 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "buffer.h"
#include "protocol.h"

#define buf_size(buf) ((buf)->end - (buf)->start)

/**
 * Allocates memory 'n' stuff.
 * @return 1 on success, 0 on error
 */
int buf_init(buffer_t *buf)
{
	buf->start = 0;
	buf->end = 0;
	buf->eof = 0;
	buf->mem_size = BUFFER_SIZE;
	buf->mem = malloc(BUFFER_SIZE);
	return buf->mem != NULL;
}

/**
 * Frees allocated memory.
 */
void buf_free(buffer_t *buf)
{
	free(buf->mem);
}

/**
 * Appends data to a buffer from a file. If buffer is full, a larger block of
 * memory is allocated for it.
 * @param fd the file descriptor
 * @param len the amount of data to be copied
 * @return the number of bytes actually read, -1 on error
 */
int buf_read_in(buffer_t *buf,
		int fd,
		int len)
{
	int size;

	if (buf->eof) {
		errno = EINVAL;
		return -1;
	}

	size = buf->end + len;
	if (size > buf->mem_size) {
		char *mem;

		mem = malloc(size);
		if (!mem) {
			return -1;
		}

		memcpy(mem, buf->mem, buf->mem_size);
		free(buf->mem);

		buf->mem_size = size;
		buf->mem = mem;
	}

	if (read_buf(fd, buf->mem + buf->end, len) < 0) {
		return -1;
	}

	buf->end += len;

	return len;
}

/**
 * Writes as much data from a buffer to a file as possible.
 * @param fd the file descriptor
 * @return the amount of data actually written, -1 on error
 */
int buf_write_out(buffer_t *buf,
		  int *fd)
{
	int size, len;

	len = 0;

	size = buf_size(buf);
	if (size) {
		if (buf->start >= buf->mem_size) {
			errno = EINVAL;
			return -1;
		}

		while (1) {
			len = write(*fd, &buf->mem[buf->start], size);
			if (len < 0) {
				if (errno == EAGAIN) {
#if 0
					continue;
#else
					len = 0;
					break;
#endif
				}
				return -1;
			}
			break;
		}

		buf->start += len;

		size = buf_size(buf);
		if (size == 0) {
			buf->start = 0;
			buf->end = 0;
		}
	}

	if (size == 0 && buf->eof) {
		close(*fd);
		*fd = -1;

		/* Make buffer empty. */
		buf->eof = 0;
	}

	return len;
}
