/*
 * $Id: daemon.c,v 1.41 2003/03/12 10:42:42 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <syslog.h>
#include <pwd.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <pty.h>
#include <utmp.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include "daemon.h"
#include "common.h"
#include "protocol.h"
#include "buffer.h"
#include "mount.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

/* The environment of the process. */
extern char **environ;

/** Debug is enabled if this is set. Used by DEBUG macro and debugfunc(). */
FILE *debugfile = NULL;

/** The server socket. */
static int srvsd = -1;

/** Socket connection to the client. */
static int sd = -1;

/** Username and password from the client. */
static auth_info_t *auth = NULL;

/** Command description from the client. */
static cmd_info_t *cmd = NULL;

/** User's uid, home dir, etc. from the system. */
static struct passwd *passwd = NULL;

/** Used when copying data between socket and the command's process. */
static buffer_t buf_in;
static char *tmp_buf = NULL;

/** Used by handler_sigchld() and manage(). */
static volatile int alive = 1;

/** Are we waiting for IN DATA? */
static int inwait = 0;

/** How much DATA has the client requested? */
static int outreq = 0;
static int errreq = 0;

/**
 * Print pid and message.
 */
void debugfunc(const char *msg, ...)
{
	va_list arg;

	fprintf(debugfile, "%5d ", getpid());

	va_start(arg, msg);
	vfprintf(debugfile, msg, arg);
	va_end(arg);

	fprintf(debugfile, "\n");
	fflush(debugfile);
}

/**
 * Prints message and errno description to syslog.
 */
static void error(const char *msg)
{
	char *desc = strerror(errno);

	if (!errno) {
		syslog(LOG_ERR, msg);
		DEBUG("Error: %s", msg);
		return;
	}

	syslog(LOG_ERR, "%s (%s)", msg, desc);
#ifdef NDEBUG
	DEBUG("Error: %s (%s)", msg, desc);
#else
	DEBUG("Error: %s (%s, %d)", msg, desc, errno);
#endif
}

/**
 * Prints message and errno description to syslog and stderr.
 */
static void error_err(const char *progname,
		      const char *msg)
{
	char *desc = strerror(errno);

	error(msg);

	if (!errno) {
		fprintf(stderr, "%s: %s", progname, msg);
		DEBUG("Error: %s", msg);
		return;
	}

	fprintf(stderr, "%s: %s (%s)\n", progname, msg, desc);
#ifdef NDEBUG
	DEBUG("Error: %s (%s)", msg, desc);
#else
	DEBUG("Error: %s (%s, %d)", msg, desc, errno);
#endif
}

/**
 * Prints message to syslog.
 */
static void warn(char *msg)
{
	syslog(LOG_WARNING, msg);
	DEBUG("Warning: %s", msg);
}

/**
 * Prints message and errno description to syslog and sends it to the client.
 */
void send_error(char *msg)
{
	int size;
	char *str, *err;

	if (!errno) {
		error(msg);
		write_packet(sd, PTYPE_ERROR, msg, strlen(msg));
		return;
	}

	err = strerror(errno);

	size = strlen(msg) + strlen(err) + 4;
	str = malloc(size);
	if (!str) {
		error(msg);
		return;
	}

	strcpy(str, msg);
	strcat(str, " (");
	strcat(str, err);
	strcat(str, ")");

	syslog(LOG_ERR, str);
	DEBUG("Error: %s", str);
	write_packet(sd, PTYPE_ERROR, str, size - 1);
}

#ifdef PTYPE_WARN
/**
 * Prints message to syslog and sends it to the client.
 */
static void send_warn(char *msg)
{
	warn(msg);
	write_packet(sd, PTYPE_WARN, msg, strlen(msg));
}
#endif

/**
 * Finds a host from a config file and returns the associated password.
 * If the host has a port defined, it will be stored in the integer pointed
 * to by portp.
 *
 * Example config file entry for daemon:
 * 192.168.0.2        foobar
 *
 * @param filename the config file path
 * @param host name of the remote machine
 * @return the password if found, NULL if not
 */
static char *find_pwd(const char *filename,
		      const char *host)
{
	FILE *file;
	char buf[1024], *buf2, *pwd = NULL;
	int len;

	file = fopen(filename, "r");
	if (!file) {
		return NULL;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}
		if (len == 0 || isspace(buf[0])) {
			continue;
		}

		buf2 = find_space(buf);
		if (*buf2 == '\0') {
			continue;
		}
		*buf2++ = '\0';

		if (strcmp(buf, host) != 0) {
			continue;
		}

		buf2 = skip_spaces(buf2);
		if (*buf2 == '\0') {
			continue;
		}
		*find_space(buf2) = '\0';

		pwd = strdup(buf2);
		break;
	}

	fclose(file);

	return pwd;
}

/**
 * Gets the amount of data we can read from a file (which may be a tty).
 */
static int get_data_length(int fd,
			   int max)
{
	int len;

	len = max;

	if (isatty(fd)) {
		if (ioctl(fd, FIONREAD, &len) < 0) {
			send_error("Can't check pty for available data");
			return -1;
		}

		if (len < 0) {
			send_error("ioctl(pty) gave invalid read length");
			return -1;
		}

		if (len > max) {
			len = max;
		}
	}

	return len;
}

/**
 * Reads some data from a fd and writes it into the socket (sd) in a packet.
 * The fd will be set to -1 at EOF.
 * @param fd pointer to the stdout or stderr of the child process
 * @param ptype the packet type
 * @return 0 if some data was copied or if the fd hit EOF,
 *         1 if no data was copied from a pty, -1 on error
 */
static int send_data(int *fd,
		     int ptype,
		     int *req)
{
	int len;

	len = get_data_length(*fd, *req);
	if (len < 0) {
		return -1;
	}
	if (len == 0) {
		return 1;
	}

	len = read(*fd, tmp_buf, len);
	if (len < 0) {
		send_error("Can't read output of the process");
		return -1;
	}

	if (len) {
		DEBUG(ptype == PTYPE_OUT_DATA ?
		      "Sending OUT DATA packet" : "Sending ERR DATA packet");

		if (write_packet(sd, ptype, tmp_buf, len) < 0) {
			error("Can't write packet to socket");
			return -1;
		}

		(*req) -= len;
	} else {
		DEBUG(ptype == PTYPE_OUT_DATA ?
		      "Stdout hit EOF" : "Stderr hit EOF");

		*fd = -1;
		*req = 0;
	}

	return 0;
}

/**
 * Sends a request for more IN DATA.
 * @return 0 on success, -1 on error
 */
static int send_request(void)
{
	DEBUG("Sending IN REQ packet");

	if (write_packet(sd, PTYPE_IN_REQ, NULL, 0) < 0) {
		error("Can't write packet to socket");
		return -1;
	}

	inwait = 1;

	return 0;
}

/**
 * Writes (some of) buf_in to a file.
 * @return 0 on success, -1 on error
 */
static int write_buffer(int *fd)
{
	DEBUG("Writing buffer to stdin");

	if (buf_write_out(&buf_in, fd) < 0) {
		send_error("Can't write buffer to stdin");
		return -1;
	}

	if (buf_is_empty(&buf_in)) {
		inwait = 0;
	}

	return 0;
}

/**
 * Reads bytes from socket to buf_in.
 * @param len the maximum number of bytes to be read
 * @return usually 0, 1 at EOF (len was zero), -1 on error
 */
static int receive_stream(int len)
{
	if (len) {
		DEBUG("Receiving IN DATA packet");

		if (buf_read_in(&buf_in, sd, len) < 0) {
			send_error("Can't append IN DATA packet to buffer");
			return -1;
		}

		return 0;
	} else {
		DEBUG("Receiving IN DATA packet: EOF");

		buf_set_eof(&buf_in);

		return 1;
	}
}

/**
 * Read an IN DATA packet from socket (sd).
 * @param fd is the pointer to the file where the data is to be copied;
 *           it will be set to -1 if IN DATA packet was empty (EOF)
 * @return 0 on success, -1 on error
 */
static int receive_packet(int *fd)
{
	phead_t head;

	if (read_phead(sd, &head) < 0) {
		error("Can't read packet header from socket");
		return -1;
	}

	switch (head.type) {
	case PTYPE_IN_DATA:
		return receive_stream(head.size);

	case PTYPE_OUT_REQ:
		DEBUG("Receiving OUT REQ packet");

		if (head.size) {
			send_error("OUT REQ packet size was non-zero");
			return -1;
		}

		outreq = BUFFER_SIZE;
		return 0;

	case PTYPE_ERR_REQ:
		DEBUG("Receiving ERR REQ packet");

		if (head.size) {
			send_error("ERR REQ packet size was non-zero");
			return -1;
		}

		errreq = BUFFER_SIZE;
		return 0;
	}

	error("Received packet has invalid type");
	return -1;
}

/**
 * Invoked at SIGCHLD signal. Sets the global alive variable to false.
 */
static void handler_sigchld(int sig)
{
	DEBUG("Handler caught SIGCHLD");
	alive = 0;
}

/**
 * Called when handler process exits.
 */
static void handler_cleanup(void)
{
	DEBUG("Handler process exiting");
}

static void print_status(int pid,
			 int status)
{
	if (WIFEXITED(status)) {
		DEBUG("%d exited normally: %d", pid, WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		DEBUG("%d was signaled: %d", pid, WTERMSIG(status));
	} else if (WIFSTOPPED(status)) {
		DEBUG("%d was stopped: %d", pid, WSTOPSIG(status));
	} else {
		DEBUG("%d died due to some other reason", pid);
	}
}

/**
 * Manages the streams of the command executor (= the child-child process).
 * @param pid of the child process, used for getting the exit status
 * @param ptyfd the file descriptor of the pseudo-tty
 */
static void manage(pid_t pid,
		   int infd,
		   int outfd,
		   int errfd)
{
	int maxfd1;
	fd_set readfds;
	fd_set writefds;
	int status = 0;
	uint16_t rc = 0;

	/* Find out the maximum fd + 1 for select(). */
	maxfd1 = MAX(sd, infd);
	maxfd1 = MAX(maxfd1, outfd);
	maxfd1 = MAX(maxfd1, errfd);
	maxfd1++;

	if (!buf_init(&buf_in)) {
		send_error("Can't initialize buffer");
		return;
	}

	/* Copy packets around until we get SIGCHLD. */
	while (alive || outfd >= 0 || errfd >= 0) {
		int copyout, copyerr;

		copyout = outfd >= 0 && (outreq || !alive);
		copyerr = errfd >= 0 && (errreq || !alive);

		FD_ZERO(&readfds);
		FD_SET(sd, &readfds);
		if (copyout) {
			FD_SET(outfd, &readfds);
		}
		if (copyerr) {
			FD_SET(errfd, &readfds);
		}

		FD_ZERO(&writefds);
		if (infd >= 0 && !(buf_is_empty(&buf_in) && inwait)) {
			FD_SET(infd, &writefds);
		}

		if (select(maxfd1, &readfds, &writefds, NULL, NULL) <= 0) {
			error("Can't select");
			return;
		}

		/* Child died and pty doesn't have anything for us? */
		if (!alive && cmd->is_tty && FD_ISSET(outfd, &readfds) &&
		    get_data_length(outfd, 1) == 0) {
			break;
		}

		/* Read REQ or DATA packet from socket. */
		if (FD_ISSET(sd, &readfds) &&
		    receive_packet(&infd) < 0) {
			return;
		}

		/* Send data from stdout. */
		if (copyout && FD_ISSET(outfd, &readfds) &&
		    send_data(&outfd, PTYPE_OUT_DATA, &outreq) < 0) {
			return;
		}

		/* Send data from stderr. */
		if (copyerr && FD_ISSET(errfd, &readfds) &&
		    send_data(&errfd, PTYPE_ERR_DATA, &errreq) < 0) {
			return;
		}

		/* Flush buf_in to stdin or request more data. */
		if (infd >= 0 && FD_ISSET(infd, &writefds)) {
			int ok;

			if (buf_is_empty(&buf_in)) {
				ok = inwait || (send_request() >= 0);
			} else {
				ok = (write_buffer(&infd) >= 0);
			}

			if (!ok) {
				return;
			}
		}
	}

	/* Get the exit status of the child process. */
	if (waitpid(pid, &status, 0) != pid) {
		send_error("Can't get exit status of the child");
		return;
	}

	rc = 1;
	if (WIFEXITED(status)) {
		rc = WEXITSTATUS(status);
	}

	print_status(pid, status);

	DEBUG("Sending RC packet");

	/* Send the return code to the client. */
	rc = htons(rc);
	if (write_packet(sd, PTYPE_RC, &rc, sizeof (rc)) < 0) {
		error("Can't write RC packet to socket");
		return;
	}
}

/**
 * @param host of the client
 * @return the truth
 */
static int authenticate(char *host)
{
	char *path, *pwd;
	int ok;

	path = malloc(strlen(passwd->pw_dir) + strlen("/" CONFIG_NAME) + 1);
	strcpy(path, passwd->pw_dir);
	strcat(path, "/" CONFIG_NAME);

	DEBUG("Searching for password in %s", path);

	pwd = find_pwd(path, host);

	free(path);

	if (!pwd) {
		warn("Can't find password from config file");
		return 0;
	}

	ok = (strcmp(pwd, auth->pwd) == 0);

	free(pwd);

	if (!ok) {
		warn("Passwords don't match");
	}
	return ok;
}

/**
 * More or less like forkpty, but the stderr of the child process is written
 * into a separate pipe.
 * @param masterptr the pty file descriptor will be written here
 * @param errfd a place to write the parent-end of the child's stderr pipe
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_pty(int *ptyfd)
{
	int master, slave, pid;
	struct winsize ws;

	ws.ws_row = cmd->tty_row;
	ws.ws_col = cmd->tty_col;
	ws.ws_xpixel = cmd->tty_xpixel;
	ws.ws_ypixel = cmd->tty_ypixel;

	if (openpty(&master, &slave, NULL, NULL, &ws) < 0) {
		send_error("Can't open a pseudo-tty");
		return -1;
	}

#if 0
	if (set_nonblocking(master) < 0) {
		send_error("Can't make pty non-blocking");
		return -1;
	}
#endif

	pid = fork();
	if (pid < 0) {
		send_error("Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		/* sd may be 0, so we must close it before login_tty(). */
		close(sd);
		sd = -1;

		close(master);
		if (login_tty(slave)) {
			send_error("Can't login to a pseudo-tty");
			exit(1);
		}
	} else {
		/* Parent */

		close(slave);
		*ptyfd = master;
	}

	return pid;
}

/**
 * More or less like fork, but the stdin/stdout/stderr of the child process
 * are set to point to sockets.
 * @param infd a place to write the parent-end of the child's stdin socketpair
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_sockets(int *infd,
			int *outfd,
			int *errfd)
{
	int insd[2], outsd[2], errsd[2], pid;

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, insd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, outsd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, errsd) < 0) {
		send_error("Can't create socket pairs");
		return -1;
	}

	if (set_nonblocking(insd[0])) {
		send_error("Can't make socket non-blocking");
		return -1;
	}

#if 0
	if (set_nonblocking(outsd[0]) < 0 ||
	    set_nonblocking(errsd[0]) < 0) {
		send_error("Can't make socket non-blocking");
		return -1;
	}
#endif

	pid = fork();
	if (pid < 0) {
		send_error("Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		/* sd may be 0, so we must close it before the dup2()s. */ 
		close(sd);
		sd = -1;

		close(insd[0]);
		close(outsd[0]);
		close(errsd[0]);

		if (dup2(insd[1], STDIN_FILENO) != STDIN_FILENO) {
			error("Can't duplicate socket as stdin");
			exit(1);
		}

		if (dup2(outsd[1], STDOUT_FILENO) != STDOUT_FILENO) {
			error("Can't duplicate socket as stdout");
			exit(1);
		}

		if (dup2(errsd[1], STDERR_FILENO) != STDERR_FILENO) {
			error("Can't duplicate socket as stderr");
			exit(1);
		}
	} else {
		/* Parent */

		close(insd[1]);
		close(outsd[1]);
		close(errsd[1]);

		*infd = insd[0];
		*outfd = outsd[0];
		*errfd = errsd[0];
	}

	return pid;
}

static int do_mount(char *str,
		    char *rootdir)
{
	mount_t *mnt;
	int yes;

	mnt = mnt_parse(str, rootdir);
	if (!mnt) {
		return -1;
	}

	yes = mnt_is_mounted(mnt);
	if (yes < 0) {
		return -1;
	}

	if (!yes) {
		DEBUG("Trying to create directory %s with uid %d",
		      mnt->point, passwd->pw_uid);

		if (seteuid(passwd->pw_uid) < 0) {
			send_error("Can't change uid");
			return -1;
		}

		if (mkdirs(mnt->point) < 0) {
			send_error("Can't create mount point");
			return -1;
		}

		if (seteuid(0) < 0) {
			send_error("Can't change uid back to root");
			return -1;
		}

		if (mnt_mount(mnt) < 0) {
			send_error("Can't mount");
			return -1;
		}
	} else {
		DEBUG("%s is already mounted", mnt->point);
	}

	mnt_free(mnt);
	return 0;
}

/**
 * Prints command line to debug log. debugfile must be defined!
 */
static void print_cmdline(void)
{
	char **p;

	fprintf(debugfile, "%5d Executing command: %s", getpid(), cmd->cmd);

	p = cmd->argv;
	if (p) {
		while (*++p) {
			fprintf(debugfile, " %s", *p);
		}
	}

	fprintf(debugfile, "\n");
	fflush(debugfile);
}

/**
 * Handles one client connection (in a child process).
 * @param host of the client
 */
static void handle(char *host)
{
	int16_t authok;
	char *rootdir, **optp;

	DEBUG("Reading auth info struct from socket");

	/* Read the authentication info. */
	auth = read_auth(sd);
	if (!auth) {
		error("Can't read authentication info");
		return;
	}

	DEBUG("Getting passwd struct of user %s", auth->user);

	/* Find out the user's uid and home. */
	passwd = getpwnam(auth->user);
	if (!passwd) {
		warn("Can't get user's passwd structure");

		/* Write negative authentication reply. */
		write_int16(sd, 0);
		return;
	}

	authok = authenticate(host);

	DEBUG("Writing auth reply to socket");

	/* Write authentication reply and return if not authenticated. */
	if (write_packet(sd, PTYPE_AUTH_OK, &authok, sizeof (authok)) < 0) {
		error("Can't write AUTH OK packet");
		return;
	}

	if (!authok) {
		return;
	}

	DEBUG("Reading command info struct from socket");

	/* Read the command info. */
	cmd = read_cmd(sd);
	if (!cmd) {
		error("Can't read command info");
		return;
	}

	/* Build path to new root. */
	rootdir = malloc(strlen(passwd->pw_dir) + 1 + strlen(cmd->target) + 1);
	if (!rootdir) {
		send_error("Out of memory\n");
		return;
	}

	strcpy(rootdir, passwd->pw_dir);
	strcat(rootdir, "/");
	strcat(rootdir, cmd->target);

	/* Mount filesystems if necessary. */
	for (optp = cmd->opts; *optp; optp++) {
		if (do_mount(*optp, rootdir) < 0) {
			return;
		}
	}

	DEBUG("Changing root directory to %s", rootdir);

	if (chroot(rootdir) < 0) {
		send_error("Can't change root directory");
		return;
	}

	free(rootdir);

	DEBUG("Changing current directory to %s", cmd->curdir);

	/* Change pwd (relative to root). */
	if (chdir("/") < 0 || chdir(cmd->curdir) < 0) {
		send_error("Can't change current directory");
		return;
	}

	DEBUG("Changing uid to %d and gid to %d",
	      passwd->pw_uid, passwd->pw_gid);

	/* Change gid to that of the user. Do this while we're still root. */
	if (setgid(passwd->pw_gid) < 0 &&
	    setegid(passwd->pw_gid) < 0) {
		send_error("Can't change group id");
		return;
	}

	/* Change uid to that of the user. */
	if (setuid(passwd->pw_uid) < 0 &&
	    seteuid(passwd->pw_uid)) {
		send_error("Can't change user id");
		return;
	}

	/* Print command line to debug log. */
	if (debugfile) {
		print_cmdline();
	}

	signal(SIGCHLD, handler_sigchld);

	if (cmd->is_tty) {
		/* TTY */

		int fd, pid;

		DEBUG("Creating command process in a pty");

		pid = fork_pty(&fd);
		if (pid < 0) {
			return;
		}

		if (pid) {
			/* Parent */
			manage(pid, fd, fd, -1);
			return;
		}
	} else {
		/* No TTY */

		int infd, outfd, errfd, pid;

		DEBUG("Creating command process without a pty");

		pid = fork_sockets(&infd, &outfd, &errfd);
		if (pid < 0) {
			return;
		}

		if (pid) {
			/* Parent */
			manage(pid, infd, outfd, errfd);
			return;
		}
	}

	/* Child */

	signal(SIGCHLD, SIG_IGN);

	environ = cmd->env;
	execvp(cmd->cmd, cmd->argv);

	error("Can't execute command");
	exit(1);
}

/**
 * Invoked at SIGCHLD signal. wait()s for the zombie child.
 */
static void daemon_sigchld(int sig)
{
	int pid, status;

	DEBUG("Daemon caught SIGCHLD");

	pid = wait(&status);
	print_status(pid, status);
}

/**
 * Called when the daemon exits.
 */
static void daemon_cleanup(void)
{
	DEBUG("Daemon process exiting");
}

static void cleanup(void)
{
	if (srvsd >= 0) {
		daemon_cleanup();
	} else if (sd >= 0) {
		handler_cleanup();
	}
}

static void print_usage(char **argv)
{
	char *name;

	name = get_progname(argv[0]);

	fprintf(stderr, "Usage: %s [port] [debugfile]\n"
		"       %s --version|-v\n", name, name);
}

/**
 * Gets the port number and debug log name. Prints usage and exits on error.
 * @return port number
 */
static int read_args(char *progname,
		     int argc,
		     char **argv)
{
	int port, i;

	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--version") == 0 ||
		    strcmp(argv[i], "-v") == 0) {
			fprintf(stderr, DAEMON_VERSION "\n");
			exit(0);
		}

		if (strcmp(argv[i], "--help") == 0 ||
		    strcmp(argv[i], "-h") == 0) {
			print_usage(argv);
			exit(0);
		}
	}

	if (argc >= 4) {
		print_usage(argv);
		exit(1);
	}

	port = DEFAULT_PORT;
	if (argc >= 2) {
		port = atoi(argv[1]);
		if (port <= 0 || port >= 65536) {
			error_err(progname, "Invalid port number");
			return 1;
		}
	}

	/* Open DEBUG log? */
	if (argc >= 3) {
		debugfile = fopen(argv[2], "a");
		if (!debugfile) {
			error_err(progname, "Can't open debug log file");
		}
	}

	return port;
}

/**
 * Parse args, open logs, create a socket, accept connections, fork handlers.
 */
int main(int argc,
	 char **argv)
{
	char *progname;
	int port;
	struct sockaddr_in srvaddr;
	pid_t pid;

	progname = get_progname(argv[0]);
	openlog(progname, LOG_PID, LOG_DAEMON);

	port = read_args(progname, argc, argv);

	/* Let's be nice. */
	atexit(cleanup);
	signal(SIGINT, exit);
	signal(SIGHUP, exit);
	signal(SIGTERM, exit);
	signal(SIGCHLD, daemon_sigchld);

	/* Create a socket, bind it and listen to it. */
	srvsd = socket(PF_INET, SOCK_STREAM, 0);
	if (srvsd < 0) {
		error_err(progname, "Can't create socket");
		return 1;
	}

	memset((char *) &srvaddr, '\0', sizeof(srvaddr));
	srvaddr.sin_family = AF_INET;
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	srvaddr.sin_port = htons(port);

	if (bind(srvsd, (struct sockaddr *) &srvaddr, sizeof (srvaddr)) < 0) {
		error_err(progname, "Can't bind socket to port");
		return 1;
	}

	if (listen(srvsd, BACKLOG) < 0) {
		error_err(progname, "Can't listen with socket");
		return 1;
	}

	/* Daemon mode; kill the parent, etc. */
	pid = fork();
	if (pid < 0) {
		error_err(progname, "Can't fork");
		return 1;
	}
	if (pid > 0) {
		printf("%d\n", pid);
		return 0;
	}

	setsid();
	chdir("/");
	umask(0);
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	/* Accept connections from clients and maintain mounts. */
	while (1) {
		struct sockaddr_in addr;
		socklen_t len;
		char host[NI_MAXHOST];
		pid_t pid;

		DEBUG("Waiting for connection");

		len = sizeof (addr);
		sd = accept(srvsd, (struct sockaddr *) &addr, &len);
		if (sd < 0) {
			error("Can't accept");
			return 1;
		}

		/* Get the IP address. */
		if (getnameinfo((struct sockaddr *) &addr, len, host,
				sizeof (host), NULL, 0, NI_NUMERICHOST) != 0) {
			send_error("Can't get client's IP address");
			continue;
		}

		DEBUG("New connection from %s", host);

		pid = fork();

		/* Handle everything in the child. */
		if (pid == 0) {
			fd_set fds;

			/* We don't need this. */
			close(srvsd);
			srvsd = -1;

			signal(SIGCHLD, SIG_IGN);

			tmp_buf = malloc(BUFFER_SIZE);
			if (!tmp_buf) {
				send_error("Out of memory");
				return 1;
			}

			handle(host);

			/* Do something until the client goes away. */
			FD_ZERO(&fds);
			FD_SET(sd, &fds);
			while (select(sd + 1, &fds, NULL, NULL, NULL) > 0 &&
			       read(sd, tmp_buf, BUFFER_SIZE) > 0);

			return 0;
		}

		if (pid < 0) {
			send_error("Can't create handler process");
		}

		close(sd);
	}

	return 0;
}
